import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from cartopy.mpl import geoaxes


def create_geoaxis(add_coastlines=True, **subplot_kw):
    """Convenience function to create a figure with a default map projection."""
    if "projection" not in subplot_kw:
        subplot_kw["projection"] = ccrs.Robinson(central_longitude=-135.58)

    _, ax = plt.subplots(subplot_kw=subplot_kw)
    ax.set_global()

    if add_coastlines:
        ax.coastlines(color="#333333", linewidth=plt.rcParams["grid.linewidth"])

    return ax


def get_current_geoaxis(**kwargs):
    """Return current axis, if it is a GeoAxes, otherwise create a new one."""
    # `plt.gcf().axes` only checks existing axes, while `plt.gca()` also creates one
    if (ax := plt.gcf().axes) and isinstance(ax[0], geoaxes.GeoAxes):
        return ax[0]
    else:
        return create_geoaxis(**kwargs)


def sample_image(arr, resampler, ax, antialias=False):
    """Sample an image in pixel-space."""
    # Get axis size in pixels
    _, _, nx, ny = np.array(ax.bbox.bounds, dtype=int)

    if antialias:
        nx *= 2
        ny *= 2

    # Create 2d-arrays for x/y coordinates in Cartopy coordinates
    xlims, ylims = ax.get_xlim(), ax.get_ylim()

    dx = (xlims[1] - xlims[0]) / nx
    dy = (ylims[1] - ylims[0]) / ny
    xvals, yvals = np.meshgrid(
        np.linspace(xlims[0] + dx / 2, xlims[1] - dx / 2, nx),
        np.linspace(ylims[0] + dy / 2, ylims[1] - dy / 2, ny),
    )

    # Transform to lat/lon coordinates
    latlon = ccrs.PlateCarree().transform_points(
        ax.projection, xvals, yvals, np.zeros_like(xvals)
    )
    valid = np.all(np.isfinite(latlon), axis=-1)

    # Fill valid points in "plot array"
    res = np.full(latlon.shape[:-1], np.nan, dtype=arr.dtype)
    res[valid] = resampler.get_values(arr, latlon[valid, :2])

    return xr.DataArray(res, coords=[("y", yvals[:, 0]), ("x", xvals[0, :])])


def map_show(
    var, resampler, ax=None, antialias=False, dpi=None, add_coastlines=True, **kwargs
):
    """Plot a variable on a Cartopy GeoAxes.

    Parameters:
        var: array-like
            Variable to plot
        resampler: Resampler
            Resampling from lon/lat coordinates to source index
        ax: cartopy.GeoAxis
        antialias: bool
            If True, sample at double the resolution for anti-aliasing
        dpi: int
            Pixel resolution of created figure
        add_coastlines: bool
            Create GeoAxes with coastlines, if none is passed
        **kwargs: Additional keyword argument passed to `plt.imshow()`
    """
    if ax is None:
        ax = get_current_geoaxis(add_coastlines=add_coastlines)

    if dpi is not None:
        ax.get_figure().set_dpi(dpi)

    img = sample_image(var, resampler, ax, antialias)
    extent = [img.x[0], img.x[-1], img.y[0], img.y[-1]]

    return ax.imshow(img, extent=extent, origin="lower", **kwargs)


def map_contour(
    var, resampler, ax=None, antialias=False, dpi=None, add_coastlines=True, **kwargs
):
    """Plot a variable on a Cartopy GeoAxes.

    Parameters:
        var: array-like
            Variable to plot
        resampler: Resampler
            Resampling from lon/lat coordinates to source index
        ax: cartopy.GeoAxis
        antialias: bool
            If True, sample at double the resolution for anti-aliasing
        dpi: int
            Pixel resolution of created figure
        add_coastlines: bool
            Create GeoAxes with coastlines, if none is passed
        **kwargs: Additional keyword argument passed to `plt.imshow()`
    """
    if ax is None:
        ax = get_current_geoaxis(add_coastlines=add_coastlines)

    if dpi is not None:
        ax.get_figure().set_dpi(dpi)

    img = sample_image(var, resampler, ax, antialias)

    return ax.contour(img.x, img.y, img, **kwargs)



import healpix as hp
import numpy as np
import xarray as xr
from scipy.spatial import Delaunay, KDTree

from . import transform


class Resampler:
    def get_values(self, m, coords):
        """Return value(s) from a map.

        Parameters:
            m: ndarray, shape (...)
            coords: ndarray, shape (N, 2)

        Returns:
            val: ndarray, shape (N)
        """
        raise NotImplementedError


class HEALPixResampler(Resampler):
    """Find the nearest cell on an HEALPix grid."""

    def __init__(self, nside, nest=True, method="nearest"):
        self.nside = nside
        self.nest = nest
        self.method = method

    def get_interp_values(self, m, coords):
        try:
            import healpy
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Install the `healpy` package to enable linear HEALPix interpolation."
            )

        if hp.nside2npix(self.nside) != m.size:
            raise ValueError(
                "Linear interpolation is only supported for full HEALPix maps!"
            )

        coords = np.asarray(coords)

        return healpy.get_interp_val(
            m,
            theta=coords[:, 0],
            phi=coords[:, 1],
            nest=self.nest,
            lonlat=True,
        )

    def get_nearest_values(self, m, coords):
        coords = np.asarray(coords)

        idx = hp.ang2pix(
            nside=self.nside,
            theta=coords[:, 0],
            phi=coords[:, 1],
            nest=self.nest,
            lonlat=True,
        )

        if isinstance(m, xr.DataArray):
            # Using coordinate **values** allows sparse HEALPix grids to be supported.
            dim = m.dims[0]
            return m.sel({dim: idx}, method="nearest").where(lambda x: x[dim] == idx)
        else:
            return m[idx]

    def get_values(self, m, coords):
        if self.method == "nearest":
            return self.get_nearest_values(m, coords)
        elif self.method == "linear":
            return self.get_interp_values(m, coords)
        else:
            raise ValueError(f"Unsupported interpolation method {self.method}.")


class KDTreeResampler(Resampler):
    """Build a KDTree for lat/lon pairs to find the nearest neighbour."""

    def __init__(self, lon, lat):
        xyz = transform.latlon2xyz(np.array([lon, lat]).T)

        self.tree = KDTree(xyz)

    def get_values(self, m, coords):
        xyz = transform.latlon2xyz(np.asarray(coords))

        _, idx = self.tree.query(xyz)

        return np.asarray(m)[idx]


class DelaunayResampler(Resampler):
    """Perform a Delaunay triangulation to find the neighbouring cells.

    References:
        https://www.redblobgames.com/x/1842-delaunay-voronoi-sphere/
    """

    def __init__(self, lon, lat):
        xy = transform.latlon2stereographic(np.array([lon, lat]).T)

        self.xyz = transform.latlon2xyz(np.array([lon, lat]).T)
        self.tri = Delaunay(xy)

    def get_values(self, m, coords):
        coords = np.asarray(coords)
        m = np.asarray(m)

        # Triangulation in stereographic projection
        xy = transform.latlon2stereographic(coords)
        triangles = self.tri.find_simplex(xy)
        valid = triangles >= 0

        idx = self.tri.simplices[triangles]

        # Compute barycentric weights in 3D
        verts_xyz = self.xyz[idx]
        v0, v1, v2 = verts_xyz[:, 0], verts_xyz[:, 1], verts_xyz[:, 2]
        p = transform.latlon2xyz(coords)

        area_total = np.linalg.norm(np.cross(v1 - v0, v2 - v0), axis=1)
        w0 = np.linalg.norm(np.cross(v1 - p, v2 - p), axis=1) / area_total
        w1 = np.linalg.norm(np.cross(v2 - p, v0 - p), axis=1) / area_total
        w2 = 1.0 - w0 - w1

        weights = np.stack([w0, w1, w2], axis=-1)

        return np.where(valid, (m[idx] * weights).sum(axis=-1), np.nan)
